"""Google Palm LLM

This module is to run the Google PaLM API hosted and maintained by Google.
To read more on Google PaLM follow
https://developers.generativeai.google/products/palm.

Example:
    Use below example to call GooglePalm Model

    >>> from pandasai.llm.google_palm import GooglePalm

Example:
    Use below example to call Google VertexAi

    >>> from pandasai.llm.google_palm import GoogleVertexai

"""
from typing import Optional

from .base import BaseGoogle


class GooglePalm(BaseGoogle):
    """Google Palm LLM
    BaseGoogle class is extended for Google Palm model. The default and only model
    support at the moment is models/text-bison-001.

    """

    model: str = "models/text-bison-001"

    def __init__(self, api_key: str, **kwargs):
        """
        __init__ method of GooglePalm Class
        Args:
            api_key (str): API Key
            **kwargs: Extended Parameters inferred from BaseGoogle Class
        """
        self._configure(api_key=api_key)
        self._set_params(**kwargs)

    def _valid_params(self):
        """Returns if the Parameters are valid or Not"""
        return super()._valid_params() + ["model"]

    def _validate(self):
        """
        A method to Validate the Model

        """

        super()._validate()

        if not self.model:
            raise ValueError("model is required.")

    def _generate_text(self, prompt: str) -> str:
        """
        Generates text for prompt

        Args:
            prompt (str): Prompt

        Returns:
            str: LLM response
        """
        self._validate()
        completion = self.genai.generate_text(
            model=self.model,
            prompt=prompt,
            temperature=self.temperature,
            top_p=self.top_p,
            top_k=self.top_k,
            max_output_tokens=self.max_output_tokens,
        )
        return completion.result

    @property
    def type(self) -> str:
        return "google-palm"


class GoogleVertexai(BaseGoogle):
    """Google Palm Vertexai LLM
    BaseGoogle class is extended for Google Palm model using Vertexai.
    The default model support at the moment is text-bison-001.
    However, user can choose to use code-bison-001 too.

    """

    def __init__(
        self, project_id: str, location: str, model: Optional[str] = None, **kwargs
    ):
        """
        A init class to implement the Google Vertexai Models

        Args:
            project_id (str): GCP project
            location (str): GCP project Location
            model Optional (str): Model to use Default to text-bison@001
            **kwargs: Arguments to control the Model Parameters
        """

        if model is None:
            self.model = "text-bison@001"
        else:
            self.model = model

        self._configurevertexai(project_id, location)
        self.project_id = project_id
        self.location = location
        self._set_params(**kwargs)

    def _valid_params(self):
        """Returns if the Parameters are valid or Not"""
        return super()._valid_params() + ["model"]

    def _validate(self):
        """
        A method to Validate the Model

        """

        super()._validate()

        if not self.model:
            raise ValueError("model is required.")

    def _generate_text(self, prompt: str) -> str:
        """
        Generates text for prompt

        Args:
            prompt (str): Prompt

        Returns:
            str: LLM response
        """
        self._validate()

        from vertexai.preview.language_models import (
            CodeGenerationModel,
            TextGenerationModel,
        )

        if self.model == "code-bison@001":
            code_generation = CodeGenerationModel.from_pretrained(self.model)

            completion = code_generation.predict(
                prefix=prompt,
                temperature=self.temperature,
                max_output_tokens=self.max_output_tokens,
            )
        else:
            text_generation = TextGenerationModel.from_pretrained(self.model)

            completion = text_generation.predict(
                prompt=prompt,
                temperature=self.temperature,
                top_p=self.top_p,
                top_k=self.top_k,
                max_output_tokens=self.max_output_tokens,
            )

        return str(completion)

    @property
    def type(self) -> str:
        return "google-vertexai"



"""
Streamlit Middleware class

Middleware to run the code in streamlit and address the issues with streamlit
integration with PandasAI.
"""

from pandasai.middlewares.base import Middleware


class StreamlitMiddleware(Middleware):
    """Streamlit Middleware class"""

    def run(self, code: str) -> str:
        """
        Run the middleware to make the code compatible with streamlit.
        For example, it replaces `plt.show()` with `st.pyplot()`.

        Returns:
            str: Modified code
        """

        code = code.replace("plt.show()", "st.pyplot(plt.gcf())")
        code = "import streamlit as st\n" + code
        return code



"""Module to import optional dependencies.

Source: Taken from pandas/compat/_optional.py
"""
from __future__ import annotations

import importlib
import sys
import warnings
from typing import TYPE_CHECKING

from pandas.util.version import Version

if TYPE_CHECKING:
    import types

# Minimum version required for each optional dependency

VERSIONS = {
    "sklearn": "1.2.2",
    "statsmodels": "0.14.0",
    "seaborn": "0.12.2",
    "plotly": "5.14.1",
    "ggplot": "0.11.5",
    "scipy": "1.9.0",
    "streamlit": "1.23.1",
}

# A mapping from import name to package name (on PyPI) for packages where
# these two names are different.

INSTALL_MAPPING = {}


def get_version(module: types.ModuleType) -> str:
    """Get the version of a module."""
    version = getattr(module, "__version__", None)

    if version is None:
        raise ImportError(f"Can't determine version for {module.__name__}")

    return version


def import_dependency(
    name: str,
    extra: str = "",
    errors: str = "raise",
    min_version: str | None = None,
):
    """
    Import an optional dependency.

    By default, if a dependency is missing an ImportError with a nice
    message will be raised. If a dependency is present, but too old,
    we raise.

    Parameters
    ----------
    name : str
        The module name.
    extra : str
        Additional text to include in the ImportError message.
    errors : str {'raise', 'warn', 'ignore'}
        What to do when a dependency is not found or its version is too old.

        * raise : Raise an ImportError
        * warn : Only applicable when a module's version is to old.
          Warns that the version is too old and returns None
        * ignore: If the module is not installed, return None, otherwise,
          return the module, even if the version is too old.
          It's expected that users validate the version locally when
          using ``errors="ignore"`` (see. ``io/html.py``)
    min_version : str, default None
        Specify a minimum version that is different from the global pandas
        minimum version required.
    Returns
    -------
    maybe_module : Optional[ModuleType]
        The imported module, when found and the version is correct.
        None is returned when the package is not found and `errors`
        is False, or when the package's version is too old and `errors`
        is ``'warn'``.
    """

    assert errors in {"warn", "raise", "ignore"}

    package_name = INSTALL_MAPPING.get(name)
    install_name = package_name if package_name is not None else name

    msg = (
        f"Missing optional dependency '{install_name}'. {extra} "
        f"Use pip or conda to install {install_name}."
    )
    try:
        module = importlib.import_module(name)
    except ImportError as exc:
        if errors == "raise":
            raise ImportError(msg) from exc
        return None

    # Handle submodules: if we have submodule, grab parent module from sys.modules
    parent = name.split(".")[0]
    if parent != name:
        install_name = parent
        module_to_get = sys.modules[install_name]
    else:
        module_to_get = module
    minimum_version = min_version if min_version is not None else VERSIONS.get(parent)
    if minimum_version:
        version = get_version(module_to_get)
        if version and Version(version) < Version(minimum_version):
            msg = (
                f"Pandas requires version '{minimum_version}' or newer of '{parent}' "
                f"(version '{version}' currently installed)."
            )
            if errors == "warn":
                warnings.warn(
                    msg,
                    UserWarning,
                )
                return None
            if errors == "raise":
                raise ImportError(msg)

    return module



""" Starcoder LLM
This module is to run the StartCoder API hosted and maintained by HuggingFace.co.
To generate HF_TOKEN go to https://huggingface.co/settings/tokens after creating Account
on the platform.

Example:
    Use below example to call Starcoder Model

    >>> from pandasai.llm.starcoder import Starcoder
"""


import os
from typing import Optional

from dotenv import load_dotenv

from ..exceptions import APIKeyNotFoundError
from .base import HuggingFaceLLM

load_dotenv()


class Starcoder(HuggingFaceLLM):

    """Starcoder LLM API

    A base HuggingFaceLLM class is extended to use Starcoder model.

    """

    api_token: str
    _api_url: str = "https://api-inference.huggingface.co/models/bigcode/starcoder"
    _max_retries: int = 5

    def __init__(self, api_token: Optional[str] = None):
        """
        __init__ method of Starcoder Class
        Args:
            api_token (str): API token from Huggingface platform
        """

        self.api_token = api_token or os.getenv("HUGGINGFACE_API_KEY") or None
        if self.api_token is None:
            raise APIKeyNotFoundError("HuggingFace Hub API key is required")

    @property
    def type(self) -> str:
        return "starcoder"



"""OpenAI LLM via Microsoft Azure Cloud

This module is to run the OpenAI API when using Microsoft Cloud infrastructure.
Azure has implemented the openai API access to its platform.
For details https://learn.microsoft.com/en-us/azure/cognitive-services/openai/reference.

Example:
    Use below example to call AzureOpenAI class

    >>> from pandasai.llm.azure_openai import AzureOpenAI

"""

import os
from typing import Any, Dict, Optional

import openai
from dotenv import load_dotenv

from ..exceptions import APIKeyNotFoundError, UnsupportedOpenAIModelError
from .base import BaseOpenAI

load_dotenv()


class AzureOpenAI(BaseOpenAI):
    """OpenAI LLM via Microsoft Azure
    This class in using BaseOpenAI class to include Azure OpenAI Features.
    """

    api_base: str
    api_type: str = "azure"
    api_version: str
    engine: str

    def __init__(
        self,
        api_token: Optional[str] = None,
        api_base: Optional[str] = None,
        api_version: Optional[str] = None,
        deployment_name: str = None,
        is_chat_model: Optional[bool] = False,
        **kwargs,
    ):
        """
        __init__ method of AzureOpenAI Class

        Args:
            api_token (str): Azure OpenAI API token.
            api_base (str): Base url of the Azure endpoint.
                It should look like the following:
                <https://YOUR_RESOURCE_NAME.openai.azure.com/>
            api_version (str): Version of the Azure OpenAI API.
                Be aware the API version may change.
            deployment_name (str): Custom name of the deployed model
            is_chat_model (bool): Whether ``deployment_name`` corresponds to a Chat
                or a Completion model
            **kwargs: Inference Parameters
        """

        self.api_token = api_token or os.getenv("OPENAI_API_KEY") or None
        self.api_base = api_base or os.getenv("OPENAI_API_BASE") or None
        self.api_version = api_version or os.getenv("OPENAI_API_VERSION")
        if self.api_token is None:
            raise APIKeyNotFoundError(
                "Azure OpenAI key is required. Please add an environment variable "
                "`OPENAI_API_KEY` or pass `api_token` as a named parameter"
            )
        if self.api_base is None:
            raise APIKeyNotFoundError(
                "Azure OpenAI base is required. Please add an environment variable "
                "`OPENAI_API_BASE` or pass `api_base` as a named parameter"
            )
        if self.api_version is None:
            raise APIKeyNotFoundError(
                "Azure OpenAI version is required. Please add an environment variable "
                "`OPENAI_API_VERSION` or pass `api_version` as a named parameter"
            )
        openai.api_key = self.api_token
        openai.api_base = self.api_base
        openai.api_version = self.api_version
        openai.api_type = self.api_type

        if deployment_name is None:
            raise UnsupportedOpenAIModelError("Model deployment name is required.")

        self.is_chat_model = is_chat_model
        self.engine = deployment_name

        self.openai_proxy = kwargs.get("openai_proxy") or os.getenv("OPENAI_PROXY")
        if self.openai_proxy:
            openai.proxy = {
                "http": self.openai_proxy,
                "https": self.openai_proxy
            }

        self._set_params(**kwargs)

    @property
    def _default_params(self) -> Dict[str, Any]:
        """Get the default parameters for calling OpenAI API

        Returns: A dict of Default Params

        """
        return {**super()._default_params, "engine": self.engine}

    def call(self, instruction: str, value: str, suffix: str = "") -> str:
        """
        Call the Azure OpenAI LLM.

        Args:
            instruction (str): Instruction to pass
            value (str): Value to pass
            suffix(str): Suffix to pass

        Returns:
            str: Response
        """
        self.last_prompt = str(instruction) + str(value)

        if self.is_chat_model:
            response = self.chat_completion(str(instruction) + str(value) + suffix)
        else:
            response = self.completion(str(instruction) + str(value) + suffix)

        return response

    @property
    def type(self) -> str:
        return "azure-openai"



""" Base class to implement a new Prompt
In order to better handle the instructions, this prompt module is written.
"""

from pandasai.exceptions import MethodNotImplementedError


class Prompt:
    """Base class to implement a new Prompt"""

    text = None
    _args = {}

    def __init__(self, **kwargs):
        """
        __init__ method of Base class of Prompt Module
        Args:
            **kwargs: Inferred Keyword Arguments
        """
        if kwargs:
            self._args = kwargs

    def __str__(self):
        if self.text is None:
            raise MethodNotImplementedError

        return self.text.format(**self._args)



""" Prompt to correct error """

import pandas as pd

from .base import Prompt


class CorrectMultipleDataframesErrorPrompt(Prompt):
    """Prompt to generate Python code"""

    text: str = """
You are provided with the the example of pandas dataframes:"""

    def __init__(
        self,
        code: str,
        error_returned: Exception,
        question: str,
        df_head: list[pd.DataFrame],
    ):
        for i, dataframe in enumerate(df_head, start=1):
            row, col = dataframe.shape
            self.text += f"""
Dataframe df{i}, with {row} rows and {col} columns.
This is the metadata of the dataframe df{i}:
{dataframe}"""

        instruction: str = f"""
The user asked the following question:
{question}

You generated this python code:
{code}

It fails with the following error:
{error_returned}

Correct the python code and return a new python code (do not import anything) that fixes the above mentioned error. Do not generate the same code again.
The natural language in the returned result must be in python comment format.
"""  # noqa: E501

        self.text += instruction

    def __str__(self):
        return self.text



from typing import Union
import pandas as pd
from abc import ABC, abstractmethod


class Shortcuts(ABC):
    @abstractmethod
    def run(self, df: pd.DataFrame, prompt: str) -> Union[str, pd.DataFrame]:
        """
        Run method from PandasAI class.

        Args:
            df (pd.DataFrame): The DataFrame containing the data.
            prompt (str): The prompt to be executed.

        Returns:
            Union[str, pd.DataFrame]: The response from the LLM.
        """

        pass

    def clean_data(self, df: pd.DataFrame) -> pd.DataFrame:
        """
        Do data cleaning and return the dataframe.

        Args:
            df (pd.DataFrame): The DataFrame containing the data.

        Returns:
            pd.DataFrame: The cleaned DataFrame.
        """

        return self.run(
            df,
            """
1. Copy the dataframe to a new variable named df_cleaned.
2. Do data cleaning.
3. Return df_cleaned.
""",
        )

    def impute_missing_values(self, df: pd.DataFrame) -> pd.DataFrame:
        """
        Do missing value imputation and return the dataframe.

        Args:
            df (pd.DataFrame): The DataFrame containing the data.

        Returns:
            pd.DataFrame: The DataFrame with imputed missing values.
        """

        return self.run(
            df,
            """
1. Copy the dataframe to a new variable named df_imputed.
2. Do the imputation of missing values.
3. Return df_imputed.
""",
        )

    def generate_features(self, df: pd.DataFrame) -> pd.DataFrame:
        """
        Do feature generation and return the dataframe.

        Args:
            df (pd.DataFrame): The DataFrame containing the data.

        Returns:
            pd.DataFrame: The DataFrame with generated features.
        """

        return self.run(
            df,
            """
1. Copy the dataframe to a new variable named df_features.
2. Do feature generation.
3. Return df_features.
""",
        )

    def plot_pie_chart(self, df: pd.DataFrame, labels: list, values: list) -> None:
        """
        Plot a pie chart.

        Args:
            df (pd.DataFrame): The DataFrame containing the data.
            labels (list): The labels for the pie chart.
            values (list): The values for the pie chart.

        Returns:
            None
        """

        self.run(
            df,
            f"""
Plot a pie chart with the following labels and values:
labels = {labels}
values = {values}
""",
        )

    def plot_bar_chart(self, df: pd.DataFrame, x: list, y: list) -> None:
        """
        Plot a bar chart.

        Args:
            df (pd.DataFrame): The DataFrame containing the data.
            x (list): The x values for the bar chart.
            y (list): The y values for the bar chart.

        Returns:
            None
        """

        self.run(
            df,
            f"""
Plot a bar chart with the following x and y:
x = {x}
y = {y}
""",
        )

    def plot_histogram(self, df: pd.DataFrame, column: str) -> None:
        """
        Plot a histogram.

        Args:
            df (pd.DataFrame): The DataFrame containing the data.
            column (str): The column to plot the histogram for.

        Returns:
            None
        """

        self.run(df, f"Plot a histogram of the column {column}.")

    def plot_line_chart(self, df: pd.DataFrame, x: list, y: list) -> None:
        """
        Plot a line chart.

        Args:
            df (pd.DataFrame): The DataFrame containing the data.
            x (list): The x values for the line chart.
            y (list): The y values for the line chart.

        Returns:
            None
        """

        self.run(
            df,
            f"""
Plot a line chart with the following x and y:
x = {x}
y = {y}
""",
        )

    def plot_scatter_chart(self, df: pd.DataFrame, x: list, y: list) -> None:
        """
        Plot a scatter chart.

        Args:
            df (pd.DataFrame): The DataFrame containing the data.
            x (list): The x values for the scatter chart.
            y (list): The y values for the scatter chart.

        Returns:
            None
        """

        self.run(
            df,
            f"""
Plot a scatter chart with the following x and y:
x = {x}
y = {y}
""",
        )

    def plot_correlation_heatmap(self, df: pd.DataFrame) -> None:
        """
        Plot a correlation heatmap.

        Args:
            df (pd.DataFrame): The DataFrame containing the data.

        Returns:
            None
        """

        self.run(df, "Plot a correlation heatmap.")

    def plot_confusion_matrix(
        self, df: pd.DataFrame, y_true: list, y_pred: list
    ) -> None:
        """
        Plot a confusion matrix.

        Args:
            df (pd.DataFrame): The DataFrame containing the data.
            y_true (list): The true values.
            y_pred (list): The predicted values.

        Returns:
            None
        """

        self.run(
            df,
            f"""
Plot a confusion matrix with the following y_true and y_pred:
y_true = {y_true}
y_pred = {y_pred}
""",
        )

    def plot_roc_curve(self, df: pd.DataFrame, y_true: list, y_pred: list) -> None:
        """
        Plot a ROC curve.

        Args:
            df (pd.DataFrame): The DataFrame containing the data.
            y_true (list): The true values.
            y_pred (list): The predicted values.

        Returns:
            None
        """

        self.run(
            df,
            f"""
Plot a ROC curve with the following y_true and y_pred:
y_true = {y_true}
y_pred = {y_pred}
""",
        )

    def boxplot(
        self,
        df: pd.DataFrame,
        col: Union[str, list[str]] = None,
        by: Union[str, list[str]] = None,
        style: str = None,
    ):
        """
        Draw a box plot to show distributions with respect to categories.

        Args:
            df (pd.DataFrame): The DataFrame containing the data.
            col (str | list[str] | None): The column(s) of interest
            for the box plot. Defaults to None.
            by (str | list[str] | None): The grouping variable(s)
            for the box plot. Defaults to None.
            style (str | None): The textual description of the desired
            style. Defaults to None.

        Returns:
            str: LLM response
        """

        if not isinstance(col, (str, list, type(None))):
            raise TypeError(
                "The 'col' argument must be a string, a list of strings, or None."
            )
        if not isinstance(by, (str, list, type(None))):
            raise TypeError(
                "The 'by' argument must be a string, a list of strings, or None."
            )

        prompt = "Plot a box-and-whisker plot"

        if isinstance(col, str):
            prompt += f" for the variable '{col}'"
        elif isinstance(col, list):
            var_list = [f"'{v}'" for v in col]
            if len(var_list) > 1:
                variables_str = ", ".join(var_list[:-1])
                prompt += f" for the variables {variables_str} and {var_list[-1]}"
            else:
                prompt += f" for the variable {var_list[0]}"

        if by is not None:
            prompt += f" grouped by '{by}'"

        if style is not None:
            prompt += f"\nStyle: '''{style}'''"

        self.run(df, prompt)

    def rolling_mean(self, df: pd.DataFrame, column: str, window: int) -> pd.DataFrame:
        """
        Calculate the rolling mean.

        Args:
            df (pd.DataFrame): The DataFrame containing the data.
            column (str): The column to calculate the rolling mean for.
            window (int): The window size.

        Returns:
            pd.DataFrame: The DataFrame containing the rolling mean.
        """

        return self.run(
            df,
            f"Calculate the rolling mean of the column {column} with a window"
            " of {window}.",
        )

    def rolling_median(
        self, df: pd.DataFrame, column: str, window: int
    ) -> pd.DataFrame:
        """
        Calculate the rolling median.

        Args:
            df (pd.DataFrame): The DataFrame containing the data.
            column (str): The column to calculate the rolling median for.
            window (int): The window size.

        Returns:
            pd.DataFrame: The DataFrame containing the rolling median.
        """

        return self.run(
            df,
            f"Calculate the rolling median of the column {column} with a window"
            " of {window}.",
        )

    def rolling_std(self, df: pd.DataFrame, column: str, window: int) -> pd.DataFrame:
        """
        Calculate the rolling standard deviation.

        Args:
            df (pd.DataFrame): The DataFrame containing the data.
            column (str): The column to calculate the rolling standard deviation for.
            window (int): The window size.

        Returns:
            pd.DataFrame: The DataFrame containing the rolling standard deviation.
        """

        return self.run(
            df,
            f"Calculate the rolling standard deviation of the column {column} with a"
            "window of {window}.",
        )

    def segment_customers(
        self, df: pd.DataFrame, features: list, n_clusters: int
    ) -> pd.DataFrame:
        """
        Segment customers.

        Args:
            df (pd.DataFrame): The DataFrame containing the data.
            features (list): The features to use for the segmentation.
            n_clusters (int): The number of clusters.

        Returns:
            pd.DataFrame: The DataFrame containing the segmentation.
        """

        return self.run(
            df,
            f"""
Segment customers with the following features and number of clusters:
features = {features}
n_clusters = {n_clusters}
""",
        )



""" Prompt to generate the response to the question in a conversational way
```
Question: {question}
Answer: {answer}

Rewrite the answer to the question in a conversational way.
```

"""

from .base import Prompt


class GenerateResponsePrompt(Prompt):
    """Prompt to generate the response to the question in a conversational way"""

    text: str = """
Question: {question}
Answer: {answer}

Rewrite the answer to the question in a conversational way.
"""



from dotenv import load_dotenv
import os
from typing import Any, Dict, Optional
from ..exceptions import APIKeyNotFoundError, UnsupportedOpenAIModelError

from langchain.chat_models import AzureChatOpenAI

from ..prompts.base import Prompt
from langchain.llms import AzureOpenAI
from .base import BaseOpenAI
from ..langchain import LangChain



load_dotenv()


class Model(BaseOpenAI):


    """AzureOpenAI LLM using BaseOpenAI Class.

    model_name: str  # gpt-35-turbo-16k  gpt-35-turbo
    api_version: str  # 2023-03-15-preview  2023-05-15
    deployment_name: str  # gpt-35-16k GPT35

    """

    api_token: str
    api_base:str
    model_name: str  
    api_version: str  
    max_tokens: int = 4097
    deployment_name: str  
    llm: AzureOpenAI
    
    _supported_chat_models = [
        "gpt-4",
        "gpt-4-0314",
        "gpt-4-32k",
        "gpt-4-32k-0314",
        "gpt-3.5-turbo",
        "gpt-3.5-turbo-0301",
    ]
    _supported_completion_models = ["text-davinci-003"]

    model: str = "gpt-3.5-turbo"
    langchain: LangChain

    def __init__(self,
        api_token: Optional[str] = None,
        api_base: Optional[str] = None,
        api_version: Optional[str] = None,
        deployment_name: str = None,
        model_name: str = None,
        **kwargs, ):

        self.api_token = api_token or os.getenv("AZURE_OPENAI_KEY") or None
        self.api_base = api_base or os.getenv("AZURE_OPENAI_ENDPOINT") or None
        self.api_version = api_version or os.getenv("AZURE_OPENAI_VERSION") or None
        self.deployment_name = deployment_name or os.getenv("DEPLOYMENT_NAME") or None
        self.model_name = model_name or os.getenv("MODEL_NAME") or None

        if self.api_token is None:
            raise APIKeyNotFoundError("Azure OpenAI key is required")
        if self.api_base is None:
            raise APIKeyNotFoundError("Azure OpenAI base endpoint is required")
        if self.deployment_name is None:
            raise UnsupportedOpenAIModelError("Model deployment name is required.")
        if self.model_name is None:
            raise UnsupportedOpenAIModelError("Model name is required.")
        if self.api_version is None:
            raise UnsupportedOpenAIModelError("Azure OpenAI version is required.")

        self.llm = AzureChatOpenAI(
            openai_api_base=self.api_base,
            openai_api_version=self.api_version,
            deployment_name=self.deployment_name,
            openai_api_key=self.api_token,
            openai_api_type="azure",
            max_tokens = self.max_tokens,
            model_name = self.model_name,
          )
        # print(self.llm)
        self.langchain = LangChain(llm = self.llm)
        self._set_params(**kwargs)

    @property
    def _default_params(self) -> Dict[str, Any]:
        """Get the default parameters for calling OpenAI API"""
        return {
            **super()._default_params,
            "model": self.model,
        }

    def call(self, instruction: Prompt, value: str, suffix: str = "", history: list = None) -> str:
        """
        Call the OpenAI LLM.

        Args:
            instruction (Prompt): Instruction to pass
            value (str): Value to pass
            suffix (str): Suffix to pass

        Raises:
            UnsupportedOpenAIModelError: Unsupported model

        Returns:
            str: Response
        """
        self.last_prompt = str(instruction) + str(value)

        if self.model in self._supported_completion_models:
            response = self.completion(str(instruction) + str(value) + suffix)
        elif self.model in self._supported_chat_models:
            # response = self.chat_completion(str(instruction) + str(value) + suffix)
            response = self.langchain_input(str(instruction) + str(value) + suffix, history)
            # print(response)
        else:
            raise UnsupportedOpenAIModelError("Unsupported model")

        return response

    @property
    def type(self) -> str:
        return "openai"
        




"""
This module contains helper functions for anonymizing data and generating random data
 before sending it to the LLM (An External API).

Only df.head() is sent to LLM API, hence the df.head() is processed
 to remove any personal or sensitive information.

"""

import random
import re
import string

import pandas as pd


def is_valid_email(email: str) -> bool:
    """Check if the given email is valid based on regex pattern.

    Args:
        email (str): email address to be checked.

    Returns (bool): True if the email is valid, otherwise False.
    """

    email_regex = r"^[a-zA-Z0-9._%+-]+@[a-zA-Z0-9.-]+\.[a-zA-Z]{2,}$"
    return re.match(email_regex, email) is not None


def is_valid_phone_number(phone_number: str) -> bool:
    """Check if the given phone number is valid based on regex pattern.

    Args:
        phone_number (str): phone number to be checked.

    Returns (bool): True if the phone number is valid, otherwise False.
    """

    pattern = r"\b(?:\+?\d{1,3}[- ]?)?\(?\d{3}\)?[- ]?\d{3}[- ]?\d{4}\b"
    return re.search(pattern, phone_number) is not None


def is_valid_credit_card(credit_card_number: str) -> bool:
    """Check if the given credit card number is valid based on regex pattern.

    Args:
        credit_card_number (str): credit card number to be checked.

    Returns (str): True if the credit card number is valid, otherwise False.
    """

    pattern = r"^\d{4}[- ]?\d{4}[- ]?\d{4}[- ]?\d{4}$"
    return re.search(pattern, credit_card_number) is not None


def generate_random_email() -> str:
    """Generates a random email address using predefined domains.

    Returns (str): generated random email address.
    """

    domains = [
        "gmail.com",
        "yahoo.com",
        "hotmail.com",
        "outlook.com",
        "icloud.com",
        "aol.com",
        "protonmail.com",
        "zoho.com",
    ]
    name_length = random.randint(6, 12)
    domain = random.choice(domains)
    letters = string.ascii_lowercase + string.digits + "-_"
    username = "".join(random.choice(letters) for i in range(name_length))
    email = username + "@" + domain
    return email


def generate_random_phone_number(original_field: str) -> str:
    """Generate a random phone number with country code if originally present.

    Args:
        original_field (str): original phone number field.

    Returns (str): generated random phone number.
    """

    if original_field.startswith("+"):
        # Extract country code if present
        country_code = original_field.split()[0]
    else:
        country_code = ""

    number = "".join(random.choices("0123456789", k=10))

    if country_code:
        phone_number = f"{country_code} {number}"
    else:
        phone_number = number

    return phone_number


def generate_random_credit_card() -> str:
    """Generate a random credit card number.

    Returns (str): generated random credit card number.
    """

    groups = []
    for _i in range(4):
        group = "".join(random.choices("0123456789", k=4))
        groups.append(group)
    separator = random.choice(["-", " "])
    return separator.join(groups)


def copy_head(data_frame: pd.DataFrame) -> pd.DataFrame:
    """Copy the head of a DataFrame.

    Args:
        data_frame (pd.DataFrame): The pd.DataFrame to copy the head from.

    Returns (pd.DataFrame): copied head of the DataFrame.
    """

    return data_frame.head().copy()


def anonymize_dataframe_head(
    data_frame: pd.DataFrame, force_conversion: bool = True
) -> pd.DataFrame:
    """Anonymize the head of a given DataFrame by replacing sensitive data.

    Args:

        data_frame (pd.DataFrame):  The DataFrame to anonymize the head data.
        force_conversion (bool): Convert it with instruction. Default is True.

    Returns: Anonymized head of the DataFrame.
    """

    data_frame = copy_head(data_frame)
    dtypes = data_frame.dtypes
    for col in data_frame.columns:
        col_idx = data_frame.columns.get_loc(col)
        # check category type column and temporarily convert to object type
        if force_conversion:
            if pd.api.types.is_categorical_dtype(data_frame[col]):
                if data_frame[col].isna().any():
                    data_frame[col] = data_frame[col].astype(object)
        for row_idx, val in enumerate(data_frame[col]):
            cell_value = str(val)

            if is_valid_email(cell_value):
                data_frame.iloc[row_idx, col_idx] = generate_random_email()
                continue
            if is_valid_phone_number(cell_value):
                data_frame.iloc[row_idx, col_idx] = generate_random_phone_number(
                    cell_value
                )
                continue
            if is_valid_credit_card(cell_value):
                data_frame.iloc[row_idx, col_idx] = generate_random_credit_card()
                continue

            # anonymize data
            random_row_index = random.choice(
                [i for i in range(len(data_frame.index)) if i != row_idx]
            )
            random_value = data_frame.iloc[random_row_index, col_idx]
            data_frame.iloc[row_idx, col_idx] = random_value
            data_frame.iloc[random_row_index, col_idx] = (
                pd.eval(cell_value) if cell_value in ["True", "False"] else cell_value
            )
    # restore the original data types
    data_frame = data_frame.astype(dtypes)
    return data_frame



""" Base class to implement a new LLM

This module is the base class to integrate the various LLMs API. This module also
includes the Base LLM classes for OpenAI, HuggingFace and Google PaLM.

Example:

    ```
    from .base import BaseOpenAI

    class CustomLLM(BaseOpenAI):

        Custom Class Starts here!!
    ```
"""

import ast
import re
from abc import ABC, abstractmethod
from typing import Any, Dict, Optional

import openai
import requests

from ..exceptions import (
    APIKeyNotFoundError,
    MethodNotImplementedError,
    NoCodeFoundError,
)
from ..helpers._optional import import_dependency
from ..prompts.base import Prompt


class LLM:
    """Base class to implement a new LLM."""

    last_prompt: Optional[str] = None

    def is_pandasai_llm(self) -> bool:
        """
        Return True if the LLM is from pandasAI.

        Returns:
            bool: True if the LLM is from pandasAI
        """
        return True

    @property
    def type(self) -> str:
        """
        Return type of LLM.

        Raises:
            APIKeyNotFoundError: Type has not been implemented

        Returns:
            str: Type of LLM a string
        """
        raise APIKeyNotFoundError("Type has not been implemented")

    def _polish_code(self, code: str) -> str:
        """
        Polish the code by removing the leading "python" or "py",  \
        removing the imports and removing trailing spaces and new lines.

        Args:
            code (str): Code

        Returns:
            str: Polished code
        """
        if re.match(r"^(python|py)", code):
            code = re.sub(r"^(python|py)", "", code)
        if re.match(r"^`.*`$", code):
            code = re.sub(r"^`(.*)`$", r"\1", code)
        code = code.strip()
        return code

    def _is_python_code(self, string):
        """
        Return True if it is valid python code.
        Args:
            string (str):

        Returns (bool): True if Python Code otherwise False

        """
        try:
            ast.parse(string)
            return True
        except SyntaxError:
            return False

    def _extract_code(self, response: str, separator: str = "```") -> str:
        """
        Extract the code from the response.

        Args:
            response (str): Response
            separator (str, optional): Separator. Defaults to "```".

        Raises:
            NoCodeFoundError: No code found in the response

        Returns:
            str: Extracted code from the response
        """
        code = response
        if len(code.split(separator)) > 1:
            code = code.split(separator)[1]
        code = self._polish_code(code)
        if not self._is_python_code(code):
            raise NoCodeFoundError("No code found in the response")

        return code

    @abstractmethod
    def call(self, instruction: Prompt, value: str, suffix: str = "") -> str:
        """
        Execute the LLM with given prompt.

        Args:
            instruction (Prompt): Prompt
            value (str): Value
            suffix (str, optional): Suffix. Defaults to "".

        Raises:
            MethodNotImplementedError: Call method has not been implemented
        """
        raise MethodNotImplementedError("Call method has not been implemented")

    def generate_code(self, instruction: Prompt, prompt: str, history: list) -> str:
        """
        Generate the code based on the instruction and the given prompt.

        Returns:
            str: Code
        """
        return self._extract_code(self.call(instruction, prompt, suffix="\n\nCode:\n", history=history))


class BaseOpenAI(LLM, ABC):
    """Base class to implement a new OpenAI LLM
    LLM base class, this class is extended to be used with OpenAI API.

    """

    api_token: str
    temperature: float = 0
    max_tokens: int = 512
    top_p: float = 1
    frequency_penalty: float = 0
    presence_penalty: float = 0.6
    stop: Optional[str] = None
    # support explicit proxy for OpenAI
    openai_proxy: Optional[str] = None

    def _set_params(self, **kwargs):
        """
        Set Parameters
        Args:
            **kwargs: ["model", "engine", "deployment_id", "temperature","max_tokens",
            "top_p", "frequency_penalty", "presence_penalty", "stop", ]

        Returns: None

        """

        valid_params = [
            "model",
            "engine",
            "deployment_id",
            "temperature",
            "max_tokens",
            "top_p",
            "frequency_penalty",
            "presence_penalty",
            "stop",
        ]
        for key, value in kwargs.items():
            if key in valid_params:
                setattr(self, key, value)

    @property
    def _default_params(self) -> Dict[str, Any]:
        """
        Get the default parameters for calling OpenAI API

        Returns (Dict): A dict of OpenAi API parameters

        """

        return {
            "temperature": self.temperature,
            "max_tokens": self.max_tokens,
            "top_p": self.top_p,
            "frequency_penalty": self.frequency_penalty,
            "presence_penalty": self.presence_penalty,
        }

    def completion(self, prompt: str) -> str:
        """
        Query the completion API

        Args:
            prompt (str): Prompt

        Returns:
            str: LLM response
        """
        params = {**self._default_params, "prompt": prompt}

        if self.stop is not None:
            params["stop"] = [self.stop]

        response = openai.Completion.create(**params)

        return response["choices"][0]["text"]

    def chat_completion(self, value: str) -> str:
        """
        Query the chat completion API

        Args:
            value (str): Prompt

        Returns:
            str: LLM response
        """
        params = {
            **self._default_params,
            "messages": [
                {
                    "role": "system",
                    "content": value,
                }
            ],
        }

        if self.stop is not None:
            params["stop"] = [self.stop]
      
        response = openai.ChatCompletion.create(**params)

        return response["choices"][0]["message"]["content"]
    
    def langchain_input(self, value:str, history:list = None) -> str:
        response = self.langchain.__call__(value, history)

        # response = "To check the data and choose a best regression model to do the regression analysis, you can start by importing the necessary libraries and reading in the dataframe df1:\n\n```python\nimport pandas as pd\nimport numpy as np\nimport matplotlib.pyplot as plt\nimport seaborn as sns\nfrom sklearn.linear_model import LinearRegression, Lasso, Ridge\nfrom sklearn.model_selection import train_test_split, cross_val_score\nfrom sklearn.metrics import r2_score, mean_squared_error\n\ndf1 = pd.read_csv('df1.csv')\n```\n\nFrom there, you can start by exploring the data and checking for any missing values:\n\n```python\ndf1.head()\ndf1.info()\ndf1.describe()\ndf1.isnull().sum()\n```\n\nNext, you can visualize the data to see if there are any trends or patterns:\n\n```python\nsns.pairplot(df1)\nplt.show()\n```\n\nAfter that, you can start building and comparing different regression models:\n\n```python\n# Split the data into training and test sets\nX_train, X_test, y_train, y_test = train_test_split(df1.drop('Daily Mean PM2.5 Concentration', axis=1),\n                                                    df1['Daily Mean PM2.5 Concentration'],\n                                                    test_size=0.2,\n                                                    random_state=42)\n\n# Fit the linear regression model and evaluate its performance\nlr = LinearRegression()\nlr.fit(X_train, y_train)\ny_pred = lr.predict(X_test)\nprint('Linear Regression R^2:', r2_score(y_test, y_pred))\nprint('Linear Regression RMSE:', np.sqrt(mean_squared_error(y_test, y_pred)))\n\n# Fit the Lasso regression model and evaluate its performance\nlasso = Lasso(alpha=0.1)\nlasso.fit(X_train, y_train)\ny_pred = lasso.predict(X_test)\nprint('Lasso Regression R^2:', r2_score(y_test, y_pred))\nprint('Lasso Regression RMSE:', np.sqrt(mean_squared_error(y_test, y_pred)))\n\n# Fit the Ridge regression model and evaluate its performance\nridge = Ridge(alpha=0.1)\nridge.fit(X_train, y_train)\ny_pred = ridge.predict(X_test)\nprint('Ridge Regression R^2:', r2_score(y_test, y_pred))\nprint('Ridge Regression RMSE:', np.sqrt(mean_squared_error(y_test, y_pred)))\n```\n\nBased on the R^2 and RMSE scores, you can choose the best regression model for your analysis."
        
        return response


class HuggingFaceLLM(LLM):
    """Base class to implement a new Hugging Face LLM.

    LLM base class is extended to be used with HuggingFace LLM Modes APIs

    """

    last_prompt: Optional[str] = None
    api_token: str
    _api_url: str = "https://api-inference.huggingface.co/models/"
    _max_retries: int = 3

    @property
    def type(self) -> str:
        return "huggingface-llm"

    def query(self, payload):
        """
        Query the HF API
        Args:
            payload: A JSON form payload

        Returns: Generated Response

        """

        headers = {"Authorization": f"Bearer {self.api_token}"}

        response = requests.post(
            self._api_url, headers=headers, json=payload, timeout=60
        )

        return response.json()[0]["generated_text"]

    def call(self, instruction: Prompt, value: str, suffix: str = "") -> str:
        """
        A call method of HuggingFaceLLM class.
        Args:
            instruction (object): A prompt object
            value (str):
            suffix (str):

        Returns (str): A string response

        """

        prompt = str(instruction)
        payload = prompt + value + suffix

        # sometimes the API doesn't return a valid response, so we retry passing the
        # output generated from the previous call as the input
        for _i in range(self._max_retries):
            response = self.query({"inputs": payload})
            payload = response
            if response.count("<endCode>") >= 2:
                break

        # replace instruction + value from the inputs to avoid showing it in the output
        output = response.replace(prompt + value + suffix, "")
        return output


class BaseGoogle(LLM):
    """Base class to implement a new Google LLM

    LLM base class is extended to be used with Google Palm API.
    """

    genai: Any
    temperature: Optional[float] = 0
    top_p: Optional[float] = 0.8
    top_k: Optional[float] = 0.3
    max_output_tokens: Optional[int] = 1000

    def _configure(self, api_key: str):
        """
        Configure Google Palm API Key
        Args:
            api_key (str): A string of API keys generated from Google Cloud

        Returns:

        """

        if not api_key:
            raise APIKeyNotFoundError("Google Palm API key is required")

        err_msg = "Install google-generativeai >= 0.1 for Google Palm API"
        genai = import_dependency("google.generativeai", extra=err_msg)

        genai.configure(api_key=api_key)
        self.genai = genai

    def _configurevertexai(self, project_id: str, location: str):
        """
        Configure Google VertexAi
        Args:
            project_id: GCP Project
            location: Location of Project

        Returns: Vertexai Object

        """

        err_msg = "Install google-cloud-aiplatform for Google Vertexai"
        vertexai = import_dependency("vertexai", extra=err_msg)
        vertexai.init(project=project_id, location=location)
        self.vertexai = vertexai

    def _valid_params(self):
        return ["temperature", "top_p", "top_k", "max_output_tokens"]

    def _set_params(self, **kwargs):
        """
        Set Parameters
        Args:
            **kwargs: ["temperature", "top_p", "top_k", "max_output_tokens"]

        Returns:

        """

        valid_params = self._valid_params()
        for key, value in kwargs.items():
            if key in valid_params:
                setattr(self, key, value)

    def _validate(self):
        """Validates the parameters for Google"""

        if self.temperature is not None and not 0 <= self.temperature <= 1:
            raise ValueError("temperature must be in the range [0.0, 1.0]")

        if self.top_p is not None and not 0 <= self.top_p <= 1:
            raise ValueError("top_p must be in the range [0.0, 1.0]")

        if self.top_k is not None and not 0 <= self.top_k <= 1:
            raise ValueError("top_k must be in the range [0.0, 1.0]")

        if self.max_output_tokens is not None and self.max_output_tokens <= 0:
            raise ValueError("max_output_tokens must be greater than zero")

    @abstractmethod
    def _generate_text(self, prompt: str) -> str:
        """
        Generates text for prompt, specific to implementation.

        Args:
            prompt (str): Prompt

        Returns:
            str: LLM response
        """
        raise MethodNotImplementedError("method has not been implemented")

    def call(self, instruction: Prompt, value: str, suffix: str = "") -> str:
        """
        Call the Google LLM.

        Args:
            instruction (object): Instruction to pass
            value (str): Value to pass
            suffix (str): Suffix to pass

        Returns:
            str: Response
        """
        self.last_prompt = str(instruction) + value
        prompt = str(instruction) + value + suffix
        return self._generate_text(prompt)



"""
Base Middleware class

Middlewares are used to modify the code before it is executed.
"""

from abc import ABC, abstractmethod

from pandasai.exceptions import MethodNotImplementedError


class Middleware(ABC):
    """Base Middleware class"""

    _has_run: str = False

    @abstractmethod
    def run(self, code: str) -> str:
        """Run the middleware"""
        raise MethodNotImplementedError

    def __call__(self, code) -> str:
        """Call the middleware"""
        self._has_run = True
        return self.run(code=code)

    @property
    def has_run(self) -> bool:
        """Return if the middleware has run"""
        return self._has_run



""" Prompt to generate Python code
```
Today is {today_date}.
You are provided with a pandas dataframe (df) with {num_rows} rows and {num_columns} columns.
This is the metadata of the dataframe:
{df_head}.

When asked about the data, your response should include a python code that describes the
dataframe `df`. Using the provided dataframe, df, return the python code to get the answer to the following question:
```
"""  # noqa: E501

from datetime import date

from .base import Prompt


class GeneratePythonCodePrompt(Prompt):
    """Prompt to generate Python code"""

    text: str = """
Today is {today_date}.
You are provided with a pandas dataframe (df) with {num_rows} rows and {num_columns} columns.
This is the metadata of the dataframe:
{df_head}.


When asked about the data, your response should include a python code(no duplicate) that describes the dataframe `df`.
Don't add too many code comments.The result must be printed at the end of the python code.If the data format is not standard, standardize and format the data first.
Using the provided dataframe, df, return the python code to get the answer to the following question:
"""  # noqa: E501

    def __init__(self, **kwargs):
        super().__init__(**kwargs, today_date=date.today())



# List of Python builtin libraries that are added to the environment by default.
WHITELISTED_BUILTINS = [
    "abs",
    "all",
    "any",
    "ascii",
    "bin",
    "bool",
    "bytearray",
    "bytes",
    "callable",
    "chr",
    "classmethod",
    "complex",
    "delattr",
    "dict",
    "dir",
    "divmod",
    "enumerate",
    "filter",
    "float",
    "format",
    "frozenset",
    "getattr",
    "hasattr",
    "hash",
    "help",
    "hex",
    "id",
    "input",
    "int",
    "isinstance",
    "issubclass",
    "iter",
    "len",
    "list",
    "locals",
    "map",
    "max",
    "memoryview",
    "min",
    "next",
    "object",
    "oct",
    "open",
    "ord",
    "pow",
    "print",
    "property",
    "range",
    "repr",
    "reversed",
    "round",
    "set",
    "setattr",
    "slice",
    "sorted",
    "staticmethod",
    "str",
    "sum",
    "super",
    "tuple",
    "type",
    "vars",
    "zip",
]

# List of Python packages that are whitelisted for import in generated code
WHITELISTED_LIBRARIES = [
    "sklearn",
    "statsmodels",
    "seaborn",
    "plotly",
    "ggplot",
    "matplotlib",
    "numpy",
    "datetime",
    "json",
    "io",
    "base64",
    "scipy",
    "streamlit",
]



""" Prompt to generate Python code for multiple dataframes """

from datetime import date

import pandas as pd

from .base import Prompt


class MultipleDataframesPrompt(Prompt):
    """Prompt to generate Python code"""

    text: str = """
Today is {today_date}.
You are provided with the following pandas dataframes:"""
    instruction: str = """
When asked about the data, your response should include a python code that describes the dataframes provided.
Don't add too many code comments.The result must be printed at the end of the python code.If the data format is not standard, standardize and format the data first.
Using the provided dataframes and no other dataframes, return the python code to get the answer to the following question:
"""  # noqa: E501

    def __init__(self, dataframes: list[pd.DataFrame]):
        for i, dataframe in enumerate(dataframes, start=1):
            row, col = dataframe.shape

            dfhead = dataframe.head(5).to_csv(index=False)

            self.text += f"""
Dataframe df{i}, with {row} rows and {col} columns.
This is the first 5 rows of the dataframe df{i}:
{dfhead}"""


        self.text += self.instruction
        self.text = self.text.format(
            today_date=date.today(),
        )

    def __str__(self):
        return self.text



""" Helper Module to Handle Jupyter Notebook
This module contains helper functions to interact with Jupyter Notebook Functionalities.

"""

from IPython import get_ipython


class Notebook:

    """Baseclass to implement Notebook helper functions"""

    def in_notebook(self) -> bool:
        """
        Checks whether the code is running inside a notebook environment.

        Returns (bool): True if the code is running inside a Jupyter notebook,
        False otherwise.
        """
        try:
            if "IPKernelApp" not in get_ipython().config:
                return False
        except (ImportError, AttributeError):
            return False
        return True

    def create_new_cell(self, contents: str) -> None:
        """
        Creates a new code cell in the Jupyter notebook and populates
        it with the specified contents.
        Args:
            contents (str): The contents to be added to the new code cell.

        ImportError:
            If the IPython module is not installed.

        AttributeError:
            If the 'get_ipython()' call raises an AttributeError, which can happen
            if the code is not running inside a Jupyter notebook.

        Returns: None

        """

        payload = {"source": "set_next_input", "text": contents, "replace": False}
        try:
            get_ipython().payload_manager.write_payload(payload, single=False)
        except (ImportError, AttributeError) as exception:
            raise exception



""" Prompt to correct Python Code on Error
```
Today is {today_date}.
You are provided with a pandas dataframe (df) with {num_rows} rows and {num_columns} columns.
This is the metadata of the dataframe:
{df_head}.

The user asked the following question:
{question}

You generated this python code:
{code}

It fails with the following error:
{error_returned}

Correct the python code and return a new python code (do not import anything) that fixes the above
mentioned error. Do not generate the same code again.
```
"""  # noqa: E501
from datetime import date

from .base import Prompt


class CorrectErrorPrompt(Prompt):
    """Prompt to Correct Python code on Error"""

    text: str = """
Today is {today_date}.
You are provided with a pandas dataframe (df) with {num_rows} rows and {num_columns} columns.
This is the metadata of the dataframe:
{df_head}.
the columns are {df_columns}.

The user asked the following question:
{question}

You generated this python code:
{code}

It fails with the following error:
{error_returned}

Correct the python code and return a new python code (do not import anything) that fixes the above mentioned error. Do not generate the same code again.
"""  # noqa: E501

    def __init__(self, **kwargs):
        super().__init__(**kwargs, today_date=date.today())



"""OpenAI LLM API

This module is to run the OpenAI API using OpenAI API.

Example:
    Use below example to call OpenAI Model

    >>> from pandasai.llm.openai import OpenAI
"""

import os
from typing import Any, Dict, Optional

import openai
from dotenv import load_dotenv

from ..exceptions import APIKeyNotFoundError, UnsupportedOpenAIModelError
from ..prompts.base import Prompt
from .base import BaseOpenAI

from ..langchain import LangChain

load_dotenv()


class OpenAI(BaseOpenAI):
    """OpenAI LLM using BaseOpenAI Class.

    An API call to OpenAi API is sent and response is recorded and returned.
    The default chat model is **gpt-3.5-turbo** while **text-davinci-003** is only
    supported completion model.
    The list of supported Chat models includes ["gpt-4", "gpt-4-0314", "gpt-4-32k",
     "gpt-4-32k-0314","gpt-3.5-turbo", "gpt-3.5-turbo-0301"].

    """

    _supported_chat_models = [
        "gpt-4",
        "gpt-4-0613",
        "gpt-4-32k",
        "gpt-4-32k-0613",
        "gpt-3.5-turbo",
        "gpt-3.5-turbo-16k",
        "gpt-3.5-turbo-0613",
        "gpt-3.5-turbo-16k-0613",
    ]
    _supported_completion_models = ["text-davinci-003"]

    model: str = "gpt-3.5-turbo"
    langchain: LangChain

    def __init__(
        self,
        api_token: Optional[str] = None,
        **kwargs,
    ):
        """
        __init__ method of OpenAI Class
        Args:
            api_token (str): API Token fro OpenAI platform.
            **kwargs: Extended Parameters inferred from BaseOpenAI class

        Returns: Response generated from OpenAI API
        """

        self.api_token = api_token or os.getenv("OPENAI_API_KEY") or None
        if self.api_token is None:
            raise APIKeyNotFoundError("OpenAI API key is required")
        openai.api_key = self.api_token

        # # langchain 
        # self.langchain = LangChain(api_token=self.api_token)
        
        self.openai_proxy = kwargs.get("openai_proxy") or os.getenv("OPENAI_PROXY")
        if self.openai_proxy:
            openai.proxy = {
                "http": self.openai_proxy,
                "https": self.openai_proxy
            }

        self._set_params(**kwargs)

    @property
    def _default_params(self) -> Dict[str, Any]:
        """Get the default parameters for calling OpenAI API"""
        return {
            **super()._default_params,
            "model": self.model,
        }

    def call(self, instruction: Prompt, value: str, suffix: str = "", history: list = None) -> str:
        """
        Call the OpenAI LLM.

        Args:
            instruction (Prompt): Instruction to pass
            value (str): Value to pass
            suffix (str): Suffix to pass

        Raises:
            UnsupportedOpenAIModelError: Unsupported model

        Returns:
            str: Response
        """
        self.last_prompt = str(instruction) + str(value)

        if self.model in self._supported_completion_models:
            response = self.completion(str(instruction) + str(value) + suffix)
        elif self.model in self._supported_chat_models:
            response = self.chat_completion(str(instruction) + str(value) + suffix)
            # response = self.langchain_input(str(instruction) + str(value) + suffix, history)
            # print(response)
        else:
            raise UnsupportedOpenAIModelError("Unsupported model")

        return response

    @property
    def type(self) -> str:
        return "openai"




# -*- coding: utf-8 -*-


from langchain.memory import ConversationBufferMemory

from langchain import OpenAI, ConversationChain, LLMChain, PromptTemplate
from langchain.memory import ConversationBufferWindowMemory

from typing import Any, Dict, Optional
from langchain.llms import AzureOpenAI




class LangChain:

    api_token: str
    conversation: ConversationChain
    llm: AzureOpenAI
    model_name: str = "gpt-3.5-turbo"
    k: int = 2
    max_tokens: int = 4097
    generated_prompt_text: str = ""
    original_conversation: ConversationChain

    template= """Assistant is a large language model trained by OpenAI.

Assistant is constantly learning and improving, and its capabilities are constantly evolving.

Assistant was smart enough to complete my request based on the incomplete dataframe I provided.Assistant can complete the prediction based on the incomplete data I provided.

Assistant can help Human to generate python code, all the replies must be coded python syntax, no natural language exists.

At the beginning and end of the code section, indicate with '```' symbol.

{history}
Human: {human_input}
Assistant:"""

    def __init__(self, llm: AzureOpenAI, **kwargs,):
        # self.llm = OpenAI(model_name=self.model_name, openai_api_key=self.api_token, temperature=0)
        if llm is None:
            raise ValueError("llm is required")
        self.llm = llm
        prompt = PromptTemplate(
            input_variables=["history","human_input"], 
            template=self.template
        )

        memory = ConversationBufferWindowMemory(k=self.k)
        # memory = ConversationBufferMemory(memory_key="history", memory_size=3)
        # memory.save_context({"input": "hi"}, {"output": "whats up"})
        # memory.save_context({"input": "not much you"}, {"output": "not much"})

        chatgpt_chain = LLMChain(
            llm=self.llm, 
            prompt=prompt, 
            verbose=True, 
            memory=memory,
        )
        self.conversation =  chatgpt_chain
        # original 
        t = """Assistant is a large language model trained by OpenAI.
{history}
Human: {human_input}
Assistant:"""
        
        p = PromptTemplate(input_variables=["history","human_input"], template=t)
        cm = ConversationBufferWindowMemory(k=2)
        ochain = LLMChain(
            llm=self.llm, 
            prompt=p, 
            verbose=True, 
            memory=cm,
        )
        self.original_conversation = ochain
        
        
    def __call__(self, prompt: str, history:list = None, **kwargs) -> str:
        if history is not None:
            self.conversation.memory.k = len(history)
            for sec in history:
                self.conversation.memory.save_context(*sec)
        self.generated_prompt_text = self.conversation.prompt.format(history=history, human_input=prompt)
        return self.conversation.predict(human_input=prompt)

    def origin_chat(self, prompt:str, history:list = None, **kwargs) -> str:
        if history is not None:
            self.original_conversation.memory.k = len(history)
            for sec in history:
                self.original_conversation.memory.save_context(*sec)
        return self.original_conversation.predict(human_input=prompt)





"""Helper functions to save charts to a file, if plt.show() is called."""
import ast
import logging
import os
from itertools import zip_longest
from os.path import dirname
from typing import Union, List

import astor


def compare_ast(
    node1: Union[ast.expr, list[ast.expr], ast.stmt, ast.AST],
    node2: Union[ast.expr, list[ast.expr], ast.stmt, ast.AST],
    ignore_args=False,
) -> bool:
    """
    Compare two AST nodes for equality.
    Source: https://stackoverflow.com/a/66733795/11080806

    Args:
        node1 (ast.AST): First AST node to compare.
        node2 (ast.AST): Second AST node to compare.
        ignore_args (bool, optional): Whether to ignore the arguments of the nodes.
            Defaults to False.

    Returns:
        bool: True if the nodes are equal, False otherwise.

    """
    if type(node1) is not type(node2):
        return False

    if isinstance(node1, ast.AST):
        for k, node in vars(node1).items():
            if k in {"lineno", "end_lineno", "col_offset", "end_col_offset", "ctx"}:
                continue
            # ignore args and keywords
            if ignore_args and (k == "args" or k == "keywords"):
                continue
            if not compare_ast(node, getattr(node2, k), ignore_args):
                return False
        return True

    if isinstance(node1, list) and isinstance(node2, list):
        return all(
            compare_ast(n1, n2, ignore_args) for n1, n2 in zip_longest(node1, node2)
        )

    return node1 == node2


def add_save_chart(
    code: str,
    folder_name: str,
    save_charts_path: str = None,
    print_save_dir: bool = True,
) -> List[str]:
    """
    Add line to code that save charts to a file, if plt.show() is called.

    Args:
        code (str): Code to add line to.
        folder_name (str): Name of folder to save charts to.
        save_charts_path (str): User Defined Path to save Charts
        print_save_dir (bool): Print the save directory to the console.
            Defaults to True.

    Returns:
        str: Code with line added.

    """

    if save_charts_path is not None:
        charts_root_dir = save_charts_path
    else:
        # define chart save directory
        charts_root_dir = dirname(dirname(dirname(__file__)))

    chart_save_dir = os.path.join(charts_root_dir, "exports", "charts", folder_name)

    tree = ast.parse(code)

    # count number of plt.show() calls
    show_count = sum(
        compare_ast(node, ast.parse("plt.show()").body[0], ignore_args=True)
        for node in ast.walk(tree)
    )

    # if there are no plt.show() calls, return the original code
    if show_count == 0:
        return [code, None]

    if not os.path.exists(chart_save_dir):
        os.makedirs(chart_save_dir)

    # iterate through the AST and add plt.savefig() calls before plt.show() calls
    counter = ord("a")
    new_body = []
    for node in tree.body:
        if compare_ast(node, ast.parse("plt.show()").body[0], ignore_args=True):
            filename = "chart"
            if show_count > 1:
                filename += f"_{chr(counter)}"
                counter += 1

            chart_save_path = os.path.join(chart_save_dir, f"{filename}.png")
            new_body.append(ast.parse(f"plt.savefig(r'{chart_save_path}')"))
        new_body.append(node)

    chart_save_msg = f"Charts saving to: {chart_save_dir}"
    if print_save_dir:
        print(chart_save_msg)
    logging.info(chart_save_msg)

    new_tree = ast.Module(body=new_body)
    code = astor.to_source(new_tree).strip()
    return [code, chart_save_path]



class Result:

    def __init__(self, status=None, error=None, content=None, picpath=None):
        self.status = status
        self.error = error
        self.content = content
        self.picpath = picpath


""" Open Assistant LLM

This module is to run the HuggingFace OpenAssistant API hosted and maintained by
HuggingFace.co. To generate HF_TOKEN go to https://huggingface.co/settings/tokens after
creating Account on the platform.

Example:
    Use below example to call OpenAssistant Model

    >>> from pandasai.llm.open_assistant import OpenAssistant
"""

import os
from typing import Optional

from dotenv import load_dotenv

from ..exceptions import APIKeyNotFoundError
from .base import HuggingFaceLLM

load_dotenv()


class OpenAssistant(HuggingFaceLLM):
    """Open Assistant LLM
    A base HuggingFaceLLM class is extended to use OpenAssistant Model via its API.
    Currently `oasst-sft-1-pythia-12b` is supported via this module.
    """

    api_token: str
    _api_url: str = (
        "https://api-inference.huggingface.co/models/"
        "OpenAssistant/oasst-sft-1-pythia-12b"
    )
    _max_retries: int = 10

    def __init__(self, api_token: Optional[str] = None):
        """
        __init__ method of OpenAssistant Class

        Raises:
            APIKeyNotFoundError: HuggingFace Hub API key is required

        Args:
            api_token (str): API token from Huggingface platform
        """
        self.api_token = api_token or os.getenv("HUGGINGFACE_API_KEY") or None
        if self.api_token is None:
            raise APIKeyNotFoundError("HuggingFace Hub API key is required")

    @property
    def type(self) -> str:
        return "open-assistant"



from pandasai.prompts.base import Prompt
from .base import LLM


class LangchainLLM(LLM):
    """
    Class to wrap Langchain LLMs and make PandasAI interoperable
    with LangChain.
    """

    _langchain_llm = None

    def __init__(self, langchain_llm):
        self._langchain_llm = langchain_llm

    def call(self, instruction: Prompt, value: str, suffix: str = "") -> str:
        prompt = str(instruction) + value + suffix
        return self._langchain_llm(prompt)

    @property
    def type(self) -> str:
        return "langchain_" + self._langchain_llm._llm_type



"""
Helper Module to Handle Data
This module collects the helper function on handling various Data Sources.

"""
from __future__ import annotations

import pandas as pd


def from_excel(file, sheet: str | int | None = 0) -> pd.DataFrame:
    """
    Return a pandas DataFrame from an Excel file.
    Wrapper for pandas.read_excel().

    Args:
        file (str):  A file path to be read
        sheet (str | int): Name of the sheet or Sheet No. Default is 0

    Returns (pd.DataFrame): A Pandas Dataframe

    """

    return pd.read_excel(file, sheet_name=sheet)



"""Cache module for caching queries."""
import glob
import os
import shelve
from pathlib import Path


class Cache:
    """Cache class for caching queries. It is used to cache queries
    to save time and money.

    Args:
        filename (str): filename to store the cache.
    """

    def __init__(self, filename="cache"):
        # define cache directory and create directory if it does not exist
        cache_dir = os.path.join(Path.cwd(), "cache")
        os.makedirs(cache_dir, mode=0o777, exist_ok=True)

        self.filepath = os.path.join(cache_dir, filename)
        self.cache = shelve.open(self.filepath)

    def set(self, key: str, value: str) -> None:
        """Set a key value pair in the cache.

        Args:
            key (str): key to store the value.
            value (str): value to store in the cache.
        """

        self.cache[key] = value

    def get(self, key: str) -> str:
        """Get a value from the cache.

        Args:
            key (str): key to get the value from the cache.

        Returns:
            str: value from the cache.
        """

        return self.cache.get(key)

    def delete(self, key: str) -> None:
        """Delete a key value pair from the cache.

        Args:
            key (str): key to delete the value from the cache.
        """

        if key in self.cache:
            del self.cache[key]

    def close(self) -> None:
        """Close the cache."""

        self.cache.close()

    def clear(self) -> None:
        """Clean the cache."""

        self.cache.clear()

    def destroy(self) -> None:
        """Destroy the cache."""
        self.cache.close()
        for cache_file in glob.glob(self.filepath + ".*"):
            os.remove(cache_file)



"""
Charts Middleware class

Middleware to handle the charts in PandasAI.
"""

from pandasai.middlewares.base import Middleware
import sys


class ChartsMiddleware(Middleware):
    """Charts Middleware class"""

    def _is_running_in_console(self) -> bool:
        """
        Check if the code is running in console or not.

        Returns:
            bool: True if running in console else False
        """

        return sys.stdout.isatty()

    def run(self, code: str) -> str:
        """
        Run the middleware to remove issues with displaying charts in PandasAI.

        Returns:
            str: Modified code
        """

        if "plt.show()" in code:
            if "plt.close('all')" not in code:
                code = code.replace("plt.show()", "plt.show()\nplt.close('all')")

            if not self._is_running_in_console():
                code = code.replace("plt.show()", "plt.show(block=False)")
        return code



# -*- coding: utf-8 -*-
"""
PandasAI is a wrapper around a LLM to make dataframes conversational

This module includes the implementation of basis  PandasAI class with methods to run
the LLMs models on Pandas dataframes. Following LLMs are implemented so far.

Example:

    This module is the Entry point of the `pandasai` package. Following is an example
    of how to use this Class.

    ```python
    import pandas as pd
    from pandasai import PandasAI

    # Sample DataFrame
    df = pd.DataFrame({
        "country": ["United States", "United Kingdom", "France", "Germany", "Italy",
        "Spain", "Canada", "Australia", "Japan", "China"],
        "gdp": [19294482071552, 2891615567872, 2411255037952, 3435817336832,
        1745433788416, 1181205135360, 1607402389504, 1490967855104, 4380756541440,
        14631844184064],
        "happiness_index": [6.94, 7.16, 6.66, 7.07, 6.38, 6.4, 7.23, 7.22, 5.87, 5.12]
    })

    # Instantiate a LLM
    from pandasai.llm.openai import OpenAI
    llm = OpenAI(api_token="YOUR_API_TOKEN")

    pandas_ai = PandasAI(llm)
    pandas_ai(df, prompt='Which are the 5 happiest countries?')

    ```
"""

import ast
import io
import logging
import re
import sys
import uuid
import time
from contextlib import redirect_stdout
from typing import List, Optional, Union, Dict, Type
import importlib.metadata

__version__ = importlib.metadata.version("idoctorai")
import astor
import pandas as pd
from .constants import (
    WHITELISTED_BUILTINS,
    WHITELISTED_LIBRARIES,
)
from .exceptions import BadImportError, LLMNotFoundError
from .helpers._optional import import_dependency
from .helpers.anonymizer import anonymize_dataframe_head
from .helpers.cache import Cache
from .helpers.notebook import Notebook
from .helpers.save_chart import add_save_chart
from .helpers.shortcuts import Shortcuts
from .llm.base import LLM
from .llm.langchain import LangchainLLM
from .middlewares.base import Middleware
from .middlewares.charts import ChartsMiddleware
from .prompts.base import Prompt
from .prompts.correct_error_prompt import CorrectErrorPrompt
from .prompts.correct_multiples_prompt import CorrectMultipleDataframesErrorPrompt
from .prompts.generate_python_code import GeneratePythonCodePrompt
from .prompts.generate_response import GenerateResponsePrompt
from .prompts.multiple_dataframes import MultipleDataframesPrompt

from pandasai.dfagent.base import create_pandas_df_agent 
from os.path import dirname
from .dfagent.result import Result


import os


def get_version():
    """
    Get the version from the package metadata
    """
    from importlib.metadata import version

    return version("idoctorai")


__version__ = get_version()


class PandasAI(Shortcuts):
    """
    PandasAI is a wrapper around a LLM to make dataframes conversational.


    This is an entry point of `pandasai` object. This class consists of methods
    to interface the LLMs with Pandas     dataframes. A pandas dataframe metadata i.e.
    df.head() and prompt is passed on to chosen LLMs API end point to generate a Python
    code to answer the questions asked. The resultant python code is run on actual data
    and answer is converted into a conversational form.

    Note:
        Do not include the `self` parameter in the ``Args`` section.
    Args:
        _llm (obj): LLMs option to be used for API access
        _verbose (bool, optional): To show the intermediate outputs e.g. python code
        generated and execution step on the prompt. Default to False
        _is_conversational_answer (bool, optional): Whether to return answer in
        conversational form. Default to False
        _enforce_privacy (bool, optional): Do not display the data on prompt in case of
        Sensitive data. Default to False
        _max_retries (int, optional): max no. of tries to generate code on failure.
        Default to 3
        _in_notebook (bool, optional): Whether to run code in notebook. Default to False
        _original_instructions (dict, optional): The dict of instruction to run. Default
        to None
        _cache (Cache, optional): Cache object to store the results. Default to None
        _enable_cache (bool, optional): Whether to enable cache. Default to True
        _logger (logging.Logger, optional): Logger object to log the messages. Default
        to None
        _logs (List[dict], optional): List of logs to be stored. Default to []
        _prompt_id (str, optional): Unique ID to differentiate calls. Default to None
        _middlewares (List[Middleware], optional): List of middlewares to run. Default
        to [ChartsMiddleware()]
        _additional_dependencies (List[dict], optional): List of additional dependencies
        to be added. Default to []
        _custom_whitelisted_dependencies (List[str], optional): List of custom
        whitelisted dependencies. Default to []
        last_code_generated (str, optional): Pass last Code if generated. Default to
        None
        last_code_executed (str, optional): Pass the last execution / run. Default to
        None
        code_output (str, optional): The code output if any. Default to None
        last_error (str, optional): Error of running code last time. Default to None
        prompt_id (str, optional): Unique ID to differentiate calls. Default to None


    Returns (str): Response to a Question related to Data

    """

    _llm: LLM
    _verbose: bool = False
    _is_conversational_answer: bool = False
    _enforce_privacy: bool = False
    _max_retries: int = 3
    _in_notebook: bool = False
    _original_instructions: dict = {
        "question": None,
        "df_head": None,
        "df_columns": None,
        "num_rows": None,
        "num_columns": None,
    }
    _cache: Cache = None
    _enable_cache: bool = True
    _prompt_id: Optional[str] = None
    _img_path: Optional[str] = None
    _middlewares: List[Middleware] = [ChartsMiddleware()]
    _additional_dependencies: List[dict] = []
    _custom_whitelisted_dependencies: List[str] = []
    _start_time: float = 0
    _enable_logging: bool = True
    _logger: logging.Logger = None
    _logs: List[str] = []
    last_code_generated: Optional[str] = None
    last_code_executed: Optional[str] = None
    code_output: Optional[str] = None
    last_error: Optional[str] = None
    _history: list = None
    _dfagent = None

    def __init__(
        self,
        llm=None,
        conversational=False,
        verbose=False,
        enforce_privacy=False,
        save_charts=False,
        save_charts_path=None,
        enable_cache=True,
        max_retries=3,
        middlewares=None,
        custom_whitelisted_dependencies=None,
        enable_logging=True,
        non_default_prompts: Optional[Dict[str, Type[Prompt]]] = None,
    ):
        """

        __init__ method of the Class PandasAI

        Args:
            llm (object): LLMs option to be used for API access. Default is None
            conversational (bool): Whether to return answer in conversational form.
            Default to False
            verbose (bool): To show the intermediate outputs e.g. python code
            generated and execution step on the prompt.  Default to False
            enforce_privacy (bool): Execute the codes with Privacy Mode ON.
            Default to False
            save_charts (bool): Save the charts generated in the notebook.
            Default to False
            enable_cache (bool): Enable the cache to store the results.
            Default to True
            middlewares (list): List of middlewares to be used. Default to None
            custom_whitelisted_dependencies (list): List of custom dependencies to
            be used. Default to None
            enable_logging (bool): Enable the logging. Default to True
            non_default_prompts (dict): Mapping from keys to replacement prompt classes.
            Used to override specific types of prompts. Defaults to None.
        """

        # configure the logging
        # noinspection PyArgumentList
        # https://stackoverflow.com/questions/61226587/pycharm-does-not-recognize-logging-basicconfig-handlers-argument
        if enable_logging:
            handlers = [logging.FileHandler("pandasai.log")]
        else:
            handlers = []

        if verbose:
            handlers.append(logging.StreamHandler(sys.stdout))
        logging.basicConfig(
            level=logging.INFO,
            format="%(asctime)s [%(levelname)s] %(message)s",
            datefmt="%Y-%m-%d %H:%M:%S",
            handlers=handlers,
        )
        self._logger = logging.getLogger(__name__)

        if llm is None:
            raise LLMNotFoundError(
                "An LLM should be provided to instantiate a PandasAI instance"
            )
        self._load_llm(llm)
        self._is_conversational_answer = conversational
        self._verbose = verbose
        self._enforce_privacy = enforce_privacy
        self._save_charts = save_charts
        self._save_charts_path = save_charts_path
        self._process_id = str(uuid.uuid4())
        self._logs = []

        self._non_default_prompts = (
            {} if non_default_prompts is None else non_default_prompts
        )

        self._max_retries = max_retries

        self.notebook = Notebook()
        self._in_notebook = self.notebook.in_notebook()

        self._enable_cache = enable_cache
        if self._enable_cache:
            self._cache = Cache()

        if middlewares is not None:
            self.add_middlewares(*middlewares)

        if custom_whitelisted_dependencies is not None:
            self._custom_whitelisted_dependencies = custom_whitelisted_dependencies

        

    def create_dfagent(self,df):
        self._dfagent = create_pandas_df_agent(self._llm.llm,df,verbose=True)

    def _load_llm(self, llm):
        """
        Check if it is a PandasAI LLM or a Langchain LLM.
        If it is a Langchain LLM, wrap it in a PandasAI LLM.

        Args:
            llm (object): LLMs option to be used for API access

        Raises:
            BadImportError: If the LLM is a Langchain LLM but the langchain package
            is not installed
        """

        try:
            llm.is_pandasai_llm()
        except AttributeError:
            llm = LangchainLLM(llm)

        self._llm = llm

    def conversational_answer(self, question: str, answer: str) -> str:
        """
        Returns the answer in conversational form about the resultant data.

        Args:
            question (str): A question in Conversational form
            answer (str): A summary / resultant Data

        Returns (str): Response

        """

        if self._enforce_privacy:
            # we don't want to send potentially sensitive data to the LLM server
            # if the user has set enforce_privacy to True
            return answer

        generate_response_instruction = self._non_default_prompts.get(
            "generate_response", GenerateResponsePrompt
        )(question=question, answer=answer)
        return self._llm.call(generate_response_instruction, "")

    def dfagent_run(self,prompt:str):
        charts_root_dir = ""
        if self._save_charts_path is not None:
            charts_root_dir = self._save_charts_path
        else:
             # define chart save directory
            charts_root_dir = dirname(dirname(dirname(__file__)))
        
        self._prompt_id = str(uuid.uuid4())
        folder_name = self._prompt_id
        chart_save_dir = os.path.join(charts_root_dir, "exports", "charts", folder_name)  

        if not os.path.exists(chart_save_dir):
            os.makedirs(chart_save_dir)
        
        filename = "chart"
        chart_save_path = os.path.join(chart_save_dir, f"{filename}.png")
         
        # chart_save_path is image save path
        tempt:str = f"""Complete the requirements of the question, and do not generate drawing code by default unless the question requires drawing.
                question:{prompt}. 
                If question requires drawing a picture, save the picture in this directory {chart_save_path}."""
        
        result = Result()
        try:
            res = self._dfagent.run(tempt)

            if os.path.exists(chart_save_path):
                result.picpath = chart_save_path
            else:
                result.picpath = None
            
            result.content = res
            result.status = 'success'
        except Exception as exception:
            result.error = (
                "Unfortunately, I was not able to answer your question, "
                "because of the following error:\n"
                f"\n{exception}\n"
            )
            result.status = 'error'
        return result

    def run(
        self,
        data_frame: Union[pd.DataFrame, List[pd.DataFrame]],
        prompt: str,
        is_conversational_answer: bool = None,
        show_code: bool = False,
        anonymize_df: bool = True,
        use_error_correction_framework: bool = True,
        history: list = None,
    ) -> Union[str, pd.DataFrame]:
        """
        Run the PandasAI to make Dataframes Conversational.

        Args:
            data_frame (Union[pd.DataFrame, List[pd.DataFrame]]): A pandas Dataframe
            prompt (str): A prompt to query about the Dataframe
            is_conversational_answer (bool): Whether to return answer in conversational
            form. Default to False
            show_code (bool): To show the intermediate python code generated on the
            prompt. Default to False
            anonymize_df (bool): Running the code with Sensitive Data. Default to True
            use_error_correction_framework (bool): Turn on Error Correction mechanism.
            Default to True

        Returns (str): Answer to the Input Questions about the DataFrame

        """

        self._start_time = time.time()

        self.log(f"Question: {prompt}")
        self.log(f"Running PandasAI with {self._llm.type} LLM...")

        self._prompt_id = str(uuid.uuid4())
        self.log(f"Prompt ID: {self._prompt_id}")

        
        self._history = history or []

        try:
            if self._enable_cache and self._cache and self._cache.get(prompt):
                self.log("Using cached response")
                code = self._cache.get(prompt)
            else:
                rows_to_display = 0 if self._enforce_privacy else 5

                multiple: bool = isinstance(data_frame, list)

                if multiple:
                    # heads = [
                    #     anonymize_dataframe_head(dataframe)
                    #     if anonymize_df
                    #     else dataframe.head(rows_to_display)
                    #     for dataframe in data_frame
                    # ]

                    heads = data_frame

                    multiple_dataframes_instruction = self._non_default_prompts.get(
                        "multiple_dataframes", MultipleDataframesPrompt
                    )
                    code = self._llm.generate_code(
                        multiple_dataframes_instruction(dataframes=heads),
                        prompt,
                        history,
                    )

                    self._original_instructions = {
                        "question": prompt,
                        "df_head": heads,
                    }

                else:
                    df_head = data_frame.head(rows_to_display)
                    if anonymize_df:
                        df_head = anonymize_dataframe_head(df_head)
                    df_head = df_head.to_csv(index=False)

                    generate_code_instruction = self._non_default_prompts.get(
                        "generate_python_code", GeneratePythonCodePrompt
                    )(
                        prompt=prompt,
                        df_head=df_head,
                        num_rows=data_frame.shape[0],
                        num_columns=data_frame.shape[1],
                    )
                    code = self._llm.generate_code(
                        generate_code_instruction,
                        prompt,
                        history,
                    )

                    self._original_instructions = {
                        "question": prompt,
                        "df_head": df_head,
                        "df_columns": data_frame.columns.tolist(),
                        "num_rows": data_frame.shape[0],
                        "num_columns": data_frame.shape[1],
                    }

                self.last_code_generated = code
                self.log(
                    f"""
                        Code generated:
                        ```
                        {code}
                        ```
                    """
                )


                if self._enable_cache and self._cache:
                    self._cache.set(prompt, code)


            if show_code and self._in_notebook:
                self.notebook.create_new_cell(code)


            for middleware in self._middlewares:
                code = middleware(code)


            answer = self.run_code(
                code,
                data_frame,
                use_error_correction_framework=use_error_correction_framework,
            )
            self.code_output = answer
            self.log(f"Answer: {answer}")

            if is_conversational_answer is None:
                is_conversational_answer = self._is_conversational_answer
            if is_conversational_answer:
                answer = self.conversational_answer(prompt, answer)
                self.log(f"Conversational answer: {answer}")

            self.log(f"Executed in: {time.time() - self._start_time}s")
            
            return answer
        except Exception as exception:
            self.last_error = str(exception)
            # print(exception)
            return (
                "Unfortunately, I was not able to answer your question, "
                "because of the following error:\n"
                f"\n{exception}\n"
            )

    def add_middlewares(self, *middlewares: List[Middleware]):
        """
        Add middlewares to PandasAI instance.

        Args:
            *middlewares: A list of middlewares

        """
        self._middlewares.extend(middlewares)


    def clear_cache(self):
        """
        Clears the cache of the PandasAI instance.
        """
        if self._cache:
            self._cache.clear()

    def __call__(
        self,
        data_frame: Union[pd.DataFrame, List[pd.DataFrame]],
        prompt: str,
        is_conversational_answer: bool = None,
        show_code: bool = False,
        anonymize_df: bool = True,
        use_error_correction_framework: bool = True,
    ) -> Union[str, pd.DataFrame]:
        """
        __call__ method of PandasAI class. It calls the `run` method.

        Args:
            data_frame:
            prompt:
            is_conversational_answer:
            show_code:
            anonymize_df:
            use_error_correction_framework:

        Returns (str): Answer to the Input Questions about the DataFrame.

        """

        return self.run(
            data_frame,
            prompt,
            is_conversational_answer,
            show_code,
            anonymize_df,
            use_error_correction_framework,
        )

    def _check_imports(self, node: Union[ast.Import, ast.ImportFrom]):
        """
        Add whitelisted imports to _additional_dependencies.

        Args:
            node (object): ast.Import or ast.ImportFrom

        Raises:
            BadImportError: If the import is not whitelisted

        """
        if isinstance(node, ast.Import):
            module = node.names[0].name
        else:
            module = node.module

        library = module.split(".")[0]

        if library == "pandas":
            return

        if library in WHITELISTED_LIBRARIES + self._custom_whitelisted_dependencies:
            for alias in node.names:
                self._additional_dependencies.append(
                    {
                        "module": module,
                        "name": alias.name,
                        "alias": alias.asname or alias.name,
                    }
                )
            return

        if library not in WHITELISTED_BUILTINS:
            raise BadImportError(library)

    def _is_df_overwrite(self, node: ast.stmt) -> bool:
        """
        Remove df declarations from the code to prevent malicious code execution.

        Args:
            node (object): ast.stmt

        Returns (bool):

        """

        return (
            isinstance(node, ast.Assign)
            and isinstance(node.targets[0], ast.Name)
            and re.match(r"df\d{0,2}$", node.targets[0].id)
        )

    def _clean_code(self, code: str) -> str:
        """
        A method to clean the code to prevent malicious code execution

        Args:
            code(str): A python code

        Returns (str): Returns a Clean Code String

        """

        tree = ast.parse(code)

        new_body = []

        # clear recent optional dependencies
        self._additional_dependencies = []

        for node in tree.body:
            if isinstance(node, (ast.Import, ast.ImportFrom)):
                self._check_imports(node)
                continue
            if self._is_df_overwrite(node):
                continue
            new_body.append(node)

        new_tree = ast.Module(body=new_body)
        return astor.to_source(new_tree).strip()

    def _get_environment(self) -> dict:
        """
        Returns the environment for the code to be executed.

        Returns (dict): A dictionary of environment variables
        """

        return {
            "pd": pd,
            **{
                lib["alias"]: getattr(import_dependency(lib["module"]), lib["name"])
                if hasattr(import_dependency(lib["module"]), lib["name"])
                else import_dependency(lib["module"])
                for lib in self._additional_dependencies
            },
            "__builtins__": {
                **{builtin: __builtins__[builtin] for builtin in WHITELISTED_BUILTINS},
            },
        }

    def _retry_run_code(self, code: str, e: Exception, multiple: bool = False):
        """
        A method to retry the code execution with error correction framework.

        Args:
            code (str): A python code
            e (Exception): An exception
            multiple (bool): A boolean to indicate if the code is for multiple
            dataframes

        Returns (str): A python code
        """

        if multiple:
            error_correcting_instruction = self._non_default_prompts.get(
                "correct_multiple_dataframes_error",
                CorrectMultipleDataframesErrorPrompt,
            )(
                code=code,
                error_returned=e,
                question=self._original_instructions["question"],
                df_head=self._original_instructions["df_head"],
            )

        else:
            error_correcting_instruction = self._non_default_prompts.get(
                "correct_error", CorrectErrorPrompt
            )(
                code=code,
                error_returned=e,
                question=self._original_instructions["question"],
                df_head=self._original_instructions["df_head"],
                num_rows=self._original_instructions["num_rows"],
                num_columns=self._original_instructions["num_columns"],
            )

        return self._llm.generate_code(error_correcting_instruction, "", self._history)

    def run_code(
        self,
        code: str,
        data_frame: pd.DataFrame,
        use_error_correction_framework: bool = True,
    ) -> str:
        """
        A method to execute the python code generated by LLMs to answer the question
        about the input dataframe. Run the code in the current context and return the
        result.

        Args:
            code (str): A python code to execute
            data_frame (pd.DataFrame): A full Pandas DataFrame
            use_error_correction_framework (bool): Turn on Error Correction mechanism.
            Default to True

        Returns (str): String representation of the result of the code execution.

        """

        multiple: bool = isinstance(data_frame, list)

        # code = "import pandas as pd\nimport matplotlib.pyplot as plt\n\ndf1 = pd.read_csv('path/to/df1.csv') # assuming data is in a csv file\ndf1['SITE_LATITUDE'] = pd.to_numeric(df1['SITE_LATITUDE']) # convert latitude column to numeric\ndf1['SITE_LONGITUDE'] = pd.to_numeric(df1['SITE_LONGITUDE']) # convert longitude column to numeric\n\nplt.scatter(df1['SITE_LONGITUDE'], df1['SITE_LATITUDE'])\nplt.xlabel('Longitude')\nplt.ylabel('Latitude')\nplt.show(block=False)\n\n"
        # code = "plt.show(block=False)\n\n"

        # Add save chart code
        if self._save_charts:
            obj = add_save_chart(
                code, self._prompt_id, self._save_charts_path, not self._verbose
            )
            # record img save path
            code = obj[0]
            self._img_path = obj[1]

        # code = "# Import libraries\nimport pandas as pd\nfrom sklearn.linear_model import LinearRegression\nfrom sklearn.model_selection import train_test_split\n\n# Load data\ndf1 = pd.read_csv('df1.csv')\n\n# Convert 'Closing Date' column to datetime and extract year\ndf1['Closing Date'] = pd.to_datetime(df1['Closing Date'])\ndf1['Year Closed'] = df1['Closing Date'].dt.year\n\n# Group data by year and count number of closures\ndf2 = df1.groupby('Year Closed')['Bank Name'].count().reset_index()\ndf2.columns = ['Year', 'Closures']\n\n# Split data into training and testing sets\nX_train, X_test, y_train, y_test = train_test_split(df2['Year'], df2['Closures'], test_size=0.2)\n\n# Fit linear regression model to training data\nreg = LinearRegression().fit(X_train.values.reshape(-1, 1), y_train)\n\n# Predict number of closures in 2024\npredicted_closures = reg.predict([[2024]])\nprint(predicted_closures)"

        # Get the code to run removing unsafe imports and df overwrites
        code_to_run = self._clean_code(code)
        self.last_code_executed = code_to_run
        self.log(
            f"""
Code running:
```
{code_to_run}
```"""
        )

        environment: dict = self._get_environment()

        if multiple:
            environment.update(
                {f"df{i}": dataframe for i, dataframe in enumerate(data_frame, start=1)}
            )
        else:
            environment["df"] = data_frame

        # Redirect standard output to a StringIO buffer
        with redirect_stdout(io.StringIO()) as output:
            count = 0
            while count < self._max_retries:
                try:
                    # Execute the code
                    exec(code_to_run, environment)
                    code = code_to_run
                    break
                except Exception as e:
                    if not use_error_correction_framework:
                        raise e

                    count += 1
                    code = self._retry_run_code(code, e, multiple)
                    # Add save chart code
                    if self._save_charts:
                        obj = add_save_chart(
                            code, self._prompt_id, self._save_charts_path, not self._verbose
                        )
                        # record img save path
                        code = obj[0]
                        self._img_path = obj[1]

                    # Get the code to run removing unsafe imports and df overwrites
                    code_to_run = self._clean_code(code)
                    self.last_code_executed = code_to_run

        captured_output = output.getvalue().strip()
        if code.count("print(") > 1:
            return captured_output

        # Evaluate the last line and return its value or the captured output
        # We do this because we want to return the right value and the right
        # type of the value. For example, if the last line is `df.head()`, we
        # want to return the head of the dataframe, not the captured output.
        lines = code.strip().split("\n")
        last_line = lines[-1].strip()

        match = re.match(r"^print\((.*)\)$", last_line)
        if match:
            last_line = match.group(1)

        try:
            result = eval(last_line, environment)

            # In some cases, the result is a tuple of values. For example, when
            # the last line is `print("Hello", "World")`, the result is a tuple
            # of two strings. In this case, we want to return a string
            if isinstance(result, tuple):
                result = " ".join([str(element) for element in result])
            
            return result
        except Exception as e:
            raise e
            # return captured_output

        
    def log(self, message: str):
        """Log a message"""
        self._logger.info(message)
        self._logs.append(message)

    @property
    def logs(self) -> List[str]:
        """Return the logs"""
        return self._logs

    def process_id(self) -> str:
        """Return the id of this PandasAI object."""
        return self._process_id

    @property
    def last_prompt_id(self) -> str:
        """Return the id of the last prompt that was run."""
        if self._prompt_id is None:
            raise ValueError("Pandas AI has not been run yet.")
        return self._prompt_id

    @property
    def last_prompt(self) -> str:
        """Return the last prompt that was executed."""
        if self._llm:
            return self._llm.last_prompt

    def output_version(self) -> str:
        version = get_version()
        return version

    def original_chat(self,prompt: str,history: list = None) -> str:
        """
        Returns the original chat between the user and the AI.

        Returns (str): Original Chat

        """
        
        answer  = self._llm.langchain.origin_chat(prompt, history)

        return answer




"""PandasAI's custom exceptions.

This module contains the implementation of Custom Exceptions.

"""


class APIKeyNotFoundError(Exception):

    """
    Raised when the API key is not defined/declared.

    Args:
        Exception (Exception): APIKeyNotFoundError
    """


class LLMNotFoundError(Exception):
    """
    Raised when the LLM is not provided.

    Args:
        Exception (Exception): LLMNotFoundError
    """


class NoCodeFoundError(Exception):
    """
    Raised when no code is found in the response.

    Args:
        Exception (Exception): NoCodeFoundError
    """


class MethodNotImplementedError(Exception):
    """
    Raised when a method is not implemented.

    Args:
        Exception (Exception): MethodNotImplementedError
    """


class UnsupportedOpenAIModelError(Exception):
    """
    Raised when an unsupported OpenAI model is used.

    Args:
        Exception (Exception): UnsupportedOpenAIModelError
    """


class BadImportError(Exception):
    """
    Raised when a library not in the whitelist is imported.

    Args:
        Exception (Exception): BadImportError
    """

    def __init__(self, library_name):
        """
        __init__ method of BadImportError Class

        Args:
            library_name (str): Name of the library that is not in the whitelist.
        """
        self.library_name = library_name
        super().__init__(
            f"Generated code includes import of {library_name} which"
            " is not in whitelist."
        )



"""Fake LLM"""

from typing import Optional

from ..prompts.base import Prompt
from .base import LLM


class FakeLLM(LLM):
    """Fake LLM"""

    _output: str = 'print("Hello world")'

    def __init__(self, output: Optional[str] = None):
        if output is not None:
            self._output = output

    def call(self, instruction: Prompt, value: str, suffix: str = "") -> str:
        self.last_prompt = str(instruction) + str(value) + suffix
        return self._output

    @property
    def type(self) -> str:
        return "fake"
